/*
/ Projektname : RFID-Reader-Modul
/ Dateiname   : RFID-Display.c
/ Version     : 1.0.0 - ATmega8
/ Datum       : 28.12.2013
/ Compiler    : WinAVR 20100110 libc 1_7_1
/ Autor       : Johannes Kretzmann
/ Copyright   : johannes.kretzmann@projektsammlung.de
/               www.projektsammlung.de
*/


#include <avr/io.h>
#include <inttypes.h>
#include <util/delay.h>
#include "RFID-Display.h"



/* -------------------------------------------------------------------------------------
   Initialisierungsroutine fr das Display 
   Parametereinstellungen: 4 Bit Modus aktivieren                   0x02
                           2 Zeilen, Zeichen 5x7                    0x28
                           Display ein, Cursor aus, kein Blinken    0x0C
                           Speicher inkrementieren, kein Scrollen   0x06
--------------------------------------------------------------------------------------*/
void InitDisplay(void)
{
    // Setze Port-Direction-Register fr DATA4 bis DATA7 sowie RS und E auf Ausgang
    DDR_DISPLAY_DATA |= ((1<<BIT_DISPLAY_DATA4) | (1<<BIT_DISPLAY_DATA5) | (1<<BIT_DISPLAY_DATA6) | (1<<BIT_DISPLAY_DATA7));
    DDR_DISPLAY_RS   |= (1<<BIT_DISPLAY_RS);
    DDR_DISPLAY_E    |= (1<<BIT_DISPLAY_E);

    _delay_us(15000);                                               // Warte 15 ms nach Power-On


    // Setze RS auf low um Daten ins Steuerregister zu schreiben
    PORT_DISPLAY_RS &= ~(1<<BIT_DISPLAY_RS);                        // Setze RS auf 0; Steuerregister


    // Initialisierung erfolgt noch mit 8 Bit
    // Function set 0x03;
    PORT_DISPLAY_DATA |=  (1<<BIT_DISPLAY_DATA4);                   // Setze DB4 auf 1
    PORT_DISPLAY_DATA |=  (1<<BIT_DISPLAY_DATA5);                   // Setze DB5 auf 1
    PORT_DISPLAY_DATA &= ~(1<<BIT_DISPLAY_DATA6);                   // Setze DB6 auf 0
    PORT_DISPLAY_DATA &= ~(1<<BIT_DISPLAY_DATA7);                   // Setze DB7 auf 0

    PORT_DISPLAY_RS   &= ~(1<<BIT_DISPLAY_RS);                      // Setze RS  auf 0


    // --- Enable-Puls fr Datenbernahme erzeugen ---
    PORT_DISPLAY_E |=  (1<<BIT_DISPLAY_E);
    asm volatile ("nop");
    PORT_DISPLAY_E &= ~(1<<BIT_DISPLAY_E);

    _delay_us(4100);


    // --- Der vorhergehende Befehl wir wiederholt gesendet ---
    // --- Enable-Puls fr Datenbername erzeugen ---
    PORT_DISPLAY_E |=  (1<<BIT_DISPLAY_E);
    asm volatile ("nop");
    PORT_DISPLAY_E &= ~ (1<<BIT_DISPLAY_E);

    _delay_us(100);                                                 // --- Warte 100 s; BusyFlag kann noch nicht geprft werden ---


    // --- Der vorhergehende Befehl wir zum dritten Mal gesendet ---
    // --- Enable-Puls fr Datenbername erzeugen ---
    PORT_DISPLAY_E |=  (1<<BIT_DISPLAY_E);
    asm volatile ("nop");
    PORT_DISPLAY_E &= ~ (1<<BIT_DISPLAY_E);

    _delay_us(100);                                                 // --- Warte 100 s; BusyFlag kann noch nicht geprft werden ---


    // Sende Kommando 4 Bit Modus zu aktivieren
    // Function set 0x02;
    PORT_DISPLAY_DATA &= ~(1<<BIT_DISPLAY_DATA4);                   // Setze DB4 auf 0
    PORT_DISPLAY_DATA |=  (1<<BIT_DISPLAY_DATA5);                   // Setze DB5 auf 1
    PORT_DISPLAY_DATA &= ~(1<<BIT_DISPLAY_DATA6);                   // Setze DB6 auf 0
    PORT_DISPLAY_DATA &= ~(1<<BIT_DISPLAY_DATA7);                   // Setze DB7 auf 0

    // --- Enable-Puls fr Datenbernahme erzeugen ---
    PORT_DISPLAY_E |=  (1<<BIT_DISPLAY_E);
    asm volatile ("nop");
    PORT_DISPLAY_E &= ~(1<<BIT_DISPLAY_E);

    _delay_us(100);


    WriteDisplayControl(Display_FUNCTION_4BIT_2LINES);              // 4Bit / 2 Zeilen / 5x7
    WriteDisplayControl(Display_ON);                                // Display ein / Cursor aus / kein Blinken
    WriteDisplayControl(Display_ENTRY_INC);                         // inkrement / kein Scrollen

    WriteDisplayControl(CLEAR_DISPLAY);

    return;
}


/* -------------------------------------------------------------------------------------
   Funktion :         Sendet einen Befehl bzw Daten im 4-Bit-Modus an das Steuer oder
                      Datenregister des Displays entsprechend der Funktion WriteDisplayControl()
                      oder WriteDisplayData() die diese Funktion aufruft.
   Parameterbergabe: Input     : data     Byte das an das Display gesendet werden soll
                      Return    : -
--------------------------------------------------------------------------------------*/
void WriteDisplay(uint8_t data)
{
    _delay_us(500);

    // lscht zuerst die Daten die an den Portpins fr das Display anliegen und schiebt dann 
    // die oberen vier Bits von Data um zwei Stellen nach Rechts, sodass die oberen vier Bits
    // deckungsgleich mit den Portpins PB2 bis PB5 sind und sendet diese an das Display
    PORT_DISPLAY_DATA &= ~((1<<BIT_DISPLAY_DATA4) | (1<<BIT_DISPLAY_DATA5) | (1<<BIT_DISPLAY_DATA6) | (1<<BIT_DISPLAY_DATA7));
    PORT_DISPLAY_DATA |= (data>>2 & 0b00111100);


    // --- Enable-Puls fr Datenbername erzeugen ---
    PORT_DISPLAY_E |=  (1<<BIT_DISPLAY_E);
    asm volatile ("nop");
    PORT_DISPLAY_E &= ~(1<<BIT_DISPLAY_E);

    // lscht zuerst die Daten die an den Portpins fr das Display anliegen und siebt dann
    // die unteren vier Bits von Data um zwei Stellen nach Links, so dass die unten vier Bits
    // deckungsgleich mit den Portpins PB2 bis PB5 sind und sendet dies an das Display
    PORT_DISPLAY_DATA &= ~((1<<BIT_DISPLAY_DATA4) | (1<<BIT_DISPLAY_DATA5) | (1<<BIT_DISPLAY_DATA6) | (1<<BIT_DISPLAY_DATA7));
    PORT_DISPLAY_DATA |= (data<<2 & 0b00111100);


    // --- Enable-Puls fr Datenbername erzeugen ---
    PORT_DISPLAY_E |=  (1<<BIT_DISPLAY_E);
    asm volatile ("nop");
    PORT_DISPLAY_E &= ~(1<<BIT_DISPLAY_E);

    return;
}


/* -------------------------------------------------------------------------------------
   Funktion :         Sendet einen Befehl im 4-Bit-Modus an das Steuerregister des Displays
   Parameterbergabe: Input     : data     Byte das an das Display gesendet werden soll
                      Return    : -
--------------------------------------------------------------------------------------*/
void WriteDisplayControl(uint8_t data)
{
    _delay_us(500);

    PORT_DISPLAY_RS &= ~(1<<BIT_DISPLAY_RS);                        // RS auf 0 setzen
    WriteDisplay(data);

    return;
}


/* -------------------------------------------------------------------------------------
   Funktion :         Sendet Daten im 4-Bit-Modus an das Datenregister des Displays
   Parameterbergabe: Input     : data     Byte das an das Display gesendet werden soll
                      Return    : -
--------------------------------------------------------------------------------------*/
void WriteDisplayData(uint8_t data)
{
    _delay_us(500);

    PORT_DISPLAY_RS |=  (1<<BIT_DISPLAY_RS);                        // RS auf 1 setzen

    WriteDisplay(data);

    return;
}


/* -------------------------------------------------------------------------------------
   Funktion :         Diese Routine lscht den am Display angezeigten Text
   Parameterbergabe: Input     : -
                      Return    : -
--------------------------------------------------------------------------------------*/
void ClearDisplay(void)
{
    WriteDisplayControl(CLEAR_DISPLAY);

    _delay_us(2000);

    return;
}


/* -------------------------------------------------------------------------------------
   Funktion :         Diese Routine positioniert den Cursor in der Zeile y (1...2)
   Parameterbergabe: Input     : y           Zeile  (1...2)
                      Return    : -
--------------------------------------------------------------------------------------*/
void SetDisplayRow(uint8_t y)
{
    uint8_t tmp; 

    if(y==1)
        tmp=0x80+0x00;                                               // 1. Zeile
    else
        tmp=0x80+0x40;                                               // 2. Zeile

    WriteDisplayControl(tmp);

    return;
}


/* -------------------------------------------------------------------------------------
   Funktion :         Diese Routine positioniert den Cursor in der Zeile y (1...2) und
                      Spalte x (0..16) 
   Parameterbergabe: Input     : y           Zeile  (1...2)
                                  x           Spalte (0..16)
                      Return    : -
--------------------------------------------------------------------------------------*/
void SetDisplayCursor(uint8_t y, uint8_t x)
{
    unsigned char tmp;

    if(y==1)
        tmp=0x80+0x00+x;     // 1. Zeile
    else
        tmp=0x80+0x40+x;     // 2. Zeile

    WriteDisplayControl(tmp);

    return;
}


/* -------------------------------------------------------------------------------------
   Funktion :         Diese Routine sendet einen String an das Dsipay
   Parameterbergabe: Input     : *data       der zu sendende Text
                      Return    : -
--------------------------------------------------------------------------------------*/
void WriteDisplayString(char *data)
{
    while(*data)
    {
        WriteDisplayData(*data);
        data++;
    }

    return;
}



