/*
/ Projektname : RFID-Reader-Modul
/ Dateiname   : RFID-Reader-Modul.c
/ Version     : 1.0.0 - ATmega8
/ Datum       : 28.12.2013 
/ Compiler    : WinAVR 20100110 libc 1_7_1
/ Autor       : Johannes Kretzmann
/ Copyright   : johannes.kretzmann@projektsammlung.de
/               www.projektsammlung.de
/
/
/
/ Mikrocontroller:                  ATmega8
/                           ---------------------- 
/                          |                      |
/                <---PC6---|1(/RESET) (ADC5/SCL)28|---PC5-->
/             RX <---PD0---|2(RXD)    (ADC4/SDA)27|---PC4-->
/             TX <---PD1---|3(TXD)        (ADC3)26|---PC3-->
/                <---PD2---|4(INT0)       (ADC2)25|---PC2-->
/  OUTPUT_U2207B <---PD3---|5(INT1)       (ADC1)24|---PC1--> DISPLAY_E
/     CFE_U2207B <---PD4---|6(CCK/T0)     (ADC0)23|---PC0--> DISPLAY_RS
/                          |7 Vcc           GND 22|
/                          |8 GND          AREF 21|
/          XTAL1 <---PB6---|9(XTAL1)       AVcc 20|
/          XTAL2 <---PB7---|10(XTAL2)      (SCK)19|---PB5--> DISPLAY_DB7
/        TASTER1 <---PD5---|11(T1)        (MISO)18|---PB4--> DISPLAY_DB6
/        TASTER2 <---PD6---|12(AIN0)  (MOSI/OC2)17|---PB3--> DISPLAY_DB5
/        TASTER3 <---PD7---|13(AIN1)  (/SS/OC1B)16|---PB2--> DISPLAY_DB4
/     LED_RELAIS <---PB0---|14(ICP1)      (OC1A)15|---PB1--> SUMMER
/                           ---------------------
/
/
/
/ RFID-Read/Write-Chip:             U2270B
/                            ------------------- 
/                           |                   |
/                      <----|1 GND     HIPASS 16|----->
/               C_PD3 <----|2 OUTPUT      RF 15|----->
/                      <----|3 /OE         VS 14|----->
/                      <----|4 INPUT  STANDBY 13|----->
/                      <----|5 MS       VBATT 12|----->
/               C_PD4 <----|6 CFE        DVS 11|----->
/                      <----|7 DGND      VEXT 10|----->
/                      <----|8 COIL2    COIL1  9|----->
/                           |                   |
/                            -------------------
/
/
/ Fuse-Bits:  --- fr externen 8 MHz Quarz ----
/             -- (1 bedeutet unprogrammiert) --
/             CKOPT/CKSEL3..1   1111            [1111] = Externer Quarz/Keramik Resonator von 3 bis 8 MHz oder grer
/                                               [0100] = Kalibrierter interner RC Oszillator 8 MHz
/             SUT1..0     11                    Quarz Oszillator, mit langsam ansteigender Spannung
/
/             --- PonyProg-Setting ---
/             CKOPT/CKSEL3..1   [ ][ ][ ][ ]   [     ] = Externer Quarz/Keramik Resonator von 3 bis 8 MHz oder grer
/                                              [x xxx] = Kalibrierter interner RC Oszillator 8 MHz
/             SUT1..0     [ ][ ]               Quarz Oszillator, mit langsam ansteigender Spannung
/
/
/ Frequenz:   8.000.000 Hz
/
/
/ Display:    Die Ansteuerung des HD44780 kompatiblen Display-Controllers
/             erfolgt im 4-Bit-Interfacemodus. Dabei sind folgende Port-Pins
/             des C mit den Pins des Display-Moduls verfunden.
/             PB2->DB4; PB3->DB5; PB4->DB6; PB5->DB7; PC0->RS; PC1->E;
/
/
/ ATmega8 
/ Ports:      LED/Relais:  PB0 -> LED_RELAIS
/             Summer:      PB1 -> SUMMER
/             U2270B:      PD3 -> OUTPUT_U2207B
/                          PD4 -> CFE_U2207B
/             Taster:      PD5 -> TASTER1
/                          PD6 -> TASTER2
/                          PD7 -> TASTER3
/             LCD-Modul:   PB2 -> DB4
/                          PB3 -> DB5
/                          PB4 -> DB6
/                          PB5 -> DB7
/                          PC0 -> RS
/                          PC1 -> E
/
/
/ U2270B-RFID
/ Ports:                OUTPUT -> C_PD3
/                       CFE    -> C_PD4
/
/
/
/
/               ----------------------------------- Beschreibungen -----------------------------------------
/ RFID-Reader:  Dieses Programm RFID-Reader ist fr den RFID-Reader Bausatz der Firma Pollin Electronic ge-
/               schrieben worden, der unter der Artikel-Nr. 810 252 auf der Webseite www.pollin.de bezogen werden
/               kann.
/               Mit diesem Programm knnen ReadOnly Transponderkarten (EM4102) die mit einer Frequenz von 125 kHz
/               arbeiten ausgelesen werden. Die Transponderkarten-Nummer wird sowohl an dem mitgelieferten Display
/               angezeigt als auch ber die serielle Schnittstelle ausgegeben. Die Ausgabe der Transponderkarten-
/               Nummer ob mit oder ohne Hersteller-IDd bzw. in dezimal oder hexadezimal kann eingestellt werden.
/               Weiterhin untersttzt dieses Programm auch die von dem Bausatz gebotene Funktion ein Relais an-
/               zusteuern und somit beispielsweise eine Trsteuerung mit Zugangsberechtigung zu realisieren. Das
/               Programm bietet hierfr die Mglichkeit vier Transponderkarten-Nummern im EEPROM abzuspeichern und
/               beim Einlesen dieser Nummern das Relais anzusteuern.
/               Mit dem Taster 1 kann das 125 KHz Signal an der Lesespule ein- bzw. ausgeschalten werden und somit
/               auch die Lesefunktion des RFID-Readers. Ob bei der Anzeige der Transponderkarten-Nummer die Her-
/               steller-ID der Transponderkarte mit angezeigt oder ausgeblendet werden soll kann ber den Taster 2
/               ein- bzw. ausgeschalten werden. Die Ausgabe der Transponderkarten-Nummer in dezimal oder hexadezimal
/               kann mit dem Taster 3 umgeschaltet werden. Die Grundeinstellung des RFID-Readers ist, dass die Lesespule
/               eingeschalten ist, die Transponderkarten-Hersteller-ID mit ausgegeben und die Transponderkarten-
/               Nummer in hexadezimal angezeigt wird.
/               Wie oben bereits erwhnt knnen bis zu vier Transponderkarten-Nummern in dem internen EEPROM abges-
/               peichert werden. Durch das gleichzeitige drcken der Taster 1 und 3 wird die zuletzt vom RFID-Reader
/               gelesene Transponderkarten-Nummer in das EEPROM geschrieben. Die Reihenfolge in der die Trasponder-
/               karten-Nummern in dem EEPROM abgespeichert werden ist wie folgt festgelegt. Nach jedem RESET wird immer
/               mit der Speicherung der Transponderkarten-Nummer auf dem ersten Speicherplatz begonnen. Die zweite
/               Transponderkaten-Nummer wird dann auf den zweiten Speicherplatz geschrieben und so weiter bis zum
/               vierten Speicherplatz. Nach der Speicherung einer Transponderkarten-Nummer im vierten Speicherplatz
/               wird bei einer weiteren Speicherung einer Transponderkarten-Nummer die Nummer die sich auf dem ersten
/               Speicherplatz befindet durch die neue Transponderkarten-Nummer berschrieben.
/
/ Achtung:      Bitte beachten Sie, dass dieses Programm nur uneingeschrnkt fr nicht kommerzielle Zwecke genutzt
/               werden darf. Ich bernehme keine Garantie fr die einwandfreie Funktion dieses Programmes und ber-
/               nehme dadurch auch keine Haftung fr Folgeschden die Aufgrund der Benutzung dieser Daten entstehen.
/               Dieses gilt auch dann, wenn ich ber einen solchen Schaden unterrichtet worden bin.
/
/
/ Zusammenfassung:       Taster 1:    RFID-Reader-Spule ein- bzw. ausschalten
/                        Taster 2:    Transponderkarten-Hersteller-ID ein- bzw. ausblenden
/                        Taster 3:    Ausgabe der Transponderkarten-Nummer in dezimal bzw. hexadezimal
/                        Taster 1/3:  Speichert die zuletzt gelesene Transponderkarten-Nummer im EEPROM
/
/
/
*/


#include <avr/io.h>
#include <string.h>
#include <inttypes.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "RFID-Display.h"
#include "RFID-U2270B.h"


#define LED_RELAIS                      0                               // Port PB0
#define SUMMER                          1                               // Port PB1
#define TASTER1                         5                               // Port PD5 - meinSpezialByte
#define TASTER2                         6                               // Port PD6 - meinSpezialByte
#define TASTER3                         7                               // Port PD7 - meinSpezialByte


#define RFID_MODUL_EIN                  0                               // meinSpezialByte
#define RFID_ID_NR_EIN                  1                               // meinSpezialByte
#define RFID_AUS_HEX_EIN                2                               // meinSpezialByte
#define LED_RELAIS_EIN                  3                               // meinSpezialByte
#define EEPROM_Tag_Gueltig              4                               // meinSpezialByte


#define F_CPU                           8000000UL                       // Systemtakt 8 MHz
#define UART_BAUD_RATE                  9600UL                          // UART-Baut-Rate
#define QBLength_TX                     28                              // TX-Pufferspeichergre


#define BASE_BIN                        2
#define BASE_DEZ                        10
#define BASE_HEX                        16

#define STRING_LENGTH_MAX               16

#define TONHOEHE_GELESEN                0x16
#define TONHOEHE_GESPEICHERT            0xF0 

#define ANZAHL_KARTEN_IDs_IM_EEPROM     4
#define RELAIS_ANZUGSZEIT               200                             // Wert x 10 ms



#define SERIELL_TRACKING




/* -------------------------------------------------------------------------------------
   Festlegung der Texte die auf dem Display angezeigt werden sollen
--------------------------------------------------------------------------------------*/
const char eeText_RFID_Reader[]     EEMEM = "RFID-Reader";
const char eeText_U2207B_Chip[]     EEMEM = "mit U2207B Chip";
const char eeText_Bereit[]          EEMEM = "Bereit.....";
const char eeText_Sende[]           EEMEM = "Lese.......";
const char eeText_Ein[]             EEMEM = "ein";
const char eeText_Aus[]             EEMEM = "aus";
const char eeText_ID_Nr[]           EEMEM = "ID-Nummer";
const char eeText_Ausgabe[]         EEMEM = "Ausgabe";
const char eeText_Hex[]             EEMEM = "Hex";
const char eeText_Dez[]             EEMEM = "Dez";
const char eeText_Leer_Zeile[]      EEMEM = "\r\n";
const char eeText_TAG_Speichern[]   EEMEM = "Speichere..";




typedef struct {    uint8_t ID;
                    uint8_t TagByte[5];
} gesTagKarten_Struct;


const gesTagKarten_Struct eegesTagKarten[ANZAHL_KARTEN_IDs_IM_EEPROM] EEMEM ={
                                                    {0x01, {0x2A, 0x00, 0x24, 0x91, 0xFD}},
                                                    {0x02, {0x4F, 0x00, 0x88, 0x59, 0xB0}},
                                                    {0x03, {0x4F, 0x00, 0x88, 0x59, 0xB1}},
                                                    {0x04, {0x4F, 0x00, 0x88, 0x59, 0xB1}},
};



/* -------------------------------------------------------------------------------------
   Deklaration und Wertefestlegung der globalen Variablen
--------------------------------------------------------------------------------------*/
volatile uint8_t  TXQueueBuffer[QBLength_TX];                           // Zwischenspeicher fr zu sendende Daten (volatile steht fr Variablen
volatile uint8_t  pInQueueTX  = 0;                                      // die in Interrupt-Funktionen manipuliert werden knnen)
volatile uint8_t  pOutQueueTX = 0;

char String[STRING_LENGTH_MAX+1];

gesTagKarten_Struct gesTagKarten;


/* -------------------------------------------------------------------------------------
   USART-TX-Interrupt-Funktion:
   Parameterbergabe: Parameter werden in globalen Variablen bergeben.
--------------------------------------------------------------------------------------*/
ISR(USART_UDRE_vect)
{
    uint8_t temp_sreg_TX;

    temp_sreg_TX = SREG;                                                // Status-Register sichern

    UDR=TXQueueBuffer [pOutQueueTX];                                    // Sendet Zeichen

    if(pOutQueueTX == QBLength_TX-1)
       pOutQueueTX = 0;
    else
       pOutQueueTX++;


    if(pOutQueueTX == pInQueueTX)
    {
       UCSRB   &= ~(1<<UDRIE);                                          // UART TX Interrupt fr leeres Sendedatenretister ausschalten
    }

    SREG = temp_sreg_TX;                                                // Status-Register rckschreiben

    return;
}



/* -------------------------------------------------------------------------------------
   Funktion :           PutStringInQueueTX()
   Parameterbergabe:   Input     : Zeichenkette
                        Return    : -
--------------------------------------------------------------------------------------*/
void PutStringInQueueTX(char *Zeichenkette)
{
    uint8_t  i;

    for(i=0; Zeichenkette[i] != 0; i++)
    {
       // Zwangspause fr den -Controller falls dieser schneller ist als die UART
       while((pInQueueTX+1 == pOutQueueTX) || ((pInQueueTX+1 == QBLength_TX-1) && (pOutQueueTX == 0)))
       { }

       TXQueueBuffer[pInQueueTX] = Zeichenkette[i];

       if(pInQueueTX == QBLength_TX-1)
          pInQueueTX = 0;
       else
          pInQueueTX++;
    }

    UCSRB |= (1<<UDRIE);                                                // UART TX Interrupt fr leeres Sendedatenretister einschalten

    return;
}




/* -------------------------------------------------------------------------------------
   Funktion :           Die Funktion Serielle_Anzeige() liest entsprechend dem bergebenen Parameter
                        den ihm zugeteilten Text aus dem EEPROM-Speicher aus und bergibt diesen der
                        Funktion PutStringInQueueTX() durch die dieser dann ber die serielle Schnitt-
                        stele bertragen wird.
   Parameterbergabe:   Input     : Anzeige_Text    bergibt den Zeiger des zu anzeigenden Textes
                        Return    : -
--------------------------------------------------------------------------------------*/
void Serielle_Anzeige(const char *Anzeige_Text)
{
    char meinHilfsString[STRING_LENGTH_MAX];

    eeprom_read_block(&meinHilfsString, Anzeige_Text, sizeof(meinHilfsString));
    PutStringInQueueTX(meinHilfsString);

    return;
}



/* -------------------------------------------------------------------------------------
   Funktion :           Die Funktion Display_Anzeige() liest entsprechend dem bergebenen Parameter
                        den ihm zugeteilten Text aus dem EEPROM-Speicher aus und bergibt diesen
                        Text der Funktion WriteDisplayString() durch die dieser dann am Display
                        angezeigt wird.
   Parameterbergabe:   Input     : Zeile           Gibt die Zeile in der der Text angezeigt wird
                                    Anzeige_Text    bergibt den Zeiger des zu anzeigenden Textes
                        Return    : -
--------------------------------------------------------------------------------------*/
void Display_Anzeige(uint8_t Zeile, const char *Anzeige_Text)
{
    char meinHilfsString[STRING_LENGTH_MAX];

    SetDisplayRow(Zeile);

    eeprom_read_block(&meinHilfsString, Anzeige_Text, sizeof(meinHilfsString));
    WriteDisplayString(meinHilfsString);

    return;
}



/* -------------------------------------------------------------------------------------
   Hauptprogramm    :   RFID-Reader-Modul (Beschreibung siehe Einleitung)
   Parameterbergabe:   Input     : -
                        Return    : -
--------------------------------------------------------------------------------------*/
int main(void)
{
    // ================= Initialisierung ===================

    // ---------- Portpinbelegung und Definition -----------

    // ------ Port B (LED_RELAIS) ------
    DDRB  |=  (1<<LED_RELAIS);                                          // setzt Port-Pin als Ausgang fr das Relais
    PORTB &=  ~(1<<LED_RELAIS);                                         // setze Port-Pin auf Low-Pegel

    // ------ Port B (Summer) ------
    DDRB  |=  (1<<SUMMER);                                              // definiert Port-Pin als Ausgang fr den Summer
    PORTB &= ~(1<<SUMMER);                                              // setze Port-Pin auf Low-Pegel

    // ------ Port D (Taster1/2 und 3) ------
    DDRD  &= ~((1<<TASTER1) | (1<<TASTER2) | (1<<TASTER3));            // definiert Port-Pins fr TASTER als Eingnge
    PORTD |=   (1<<TASTER1) | (1<<TASTER2) | (1<<TASTER3);             // aktiviere internen Pull-Up-Widerstand vom Port-Pin


    // ---------------------------- Timer 1 (ATmega8) --------------------------
    // --- Timer 1 ist ein 16-Bit Timer und wird hier im CTC-Modus verwendet ---
    // non-PWM mode; Normal port operation; CPU-Clock/64 prescaling
    // TCCR1A: (0<<COM1A1)|(0<<COM1A0)|(0<<COM1B1)|(0<<COM1B0)|(0<<FOC1A)|(0<<FOC1B)|(0<<WGM11)|(0<<WGM10)
    // TCCR1B: (0<<ICNC1) |(0<<ICES1) |(-)|(WGM13)|(0<<WGM12)|(0<<CS12)|(1<<CS11)|(1<<CS10)
    //
    // WGM13=0, WGM12=1, WGM11=0, WGM10=0                                // CTC-Modus aktiviert 
    // COM1B1=0 ,COM1B0=0                                                // die Pin-Funktion des Port-Ausgangs fr OC1A Toggle aktivieren
    // CS12=0 , CS11=1 ; CS10=1                                          // Vorteiler 64      
    // TCCR1A = 0b01000000;                                              // Ton an, da COM1A1=0 und COM1A0=1 die Pin-Funktion des Port-Ausgangs fr OC1A Toggle aktivieren
       TCCR1A = 0b00000000;                                              // Ton aus, da COM1A1=0 und COM1A=0 die Pin-Funktion fr den CTC-Modus deaktivieren
       TCCR1B = 0b00001011;

       OCR1AH=0x00;                                                      // Festlegung der Tonhhe im OCR1A-Register
       OCR1AL=TONHOEHE_GELESEN;                                          // Festlegung der Tonhhe im OCR1A-Register



    // -------- Timer0 fr interne Zeitmessung ---------                 // Timer0 ist ein 8 Bit Timer
    // Parameterfestlegung siehe InitRFID(void)


    // -------------------- UART --------------------
    uint16_t  myUBRR;                                                    // 16-Bit breite Speichervariable fr Bautratenberechnung 
    myUBRR  = F_CPU/(UART_BAUD_RATE*16L)-1;                             // Berechnung der Bautratenwerte fr das UBRR-Register im Asynchronous Norma
    UBRRH   = (uint8_t) (myUBRR>>8);                                    // Setzt Highbyte des UBRR-REgisters
    UBRRL   = (uint8_t) myUBRR;                                         // Setzt Lowbyte  des UBRR-Registers
    UCSRB   = (1<<RXCIE) | (1<<RXEN) | (1<<TXEN);                      // UART TX und RX einschalten und RX Interrupt aktivieren 
    UCSRC   = (1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0);                    // Setzt die Bits fr die Betriebsart Asynchron 8N1 	



    sei();                                                              // aktiviert global alle definierten Interrupte

    InitDisplay();
    InitRFID();


    // ------------ lokale Variablen ------------
    uint8_t i;
    uint8_t Base;
    uint8_t Laenge;
    uint8_t Ergebnis;
    uint8_t TagByteNr;
    uint8_t Anzeigedauer;
    uint8_t AnfangsTagByteNr;
    uint8_t meinSpezialByte;                                             // Temporrer Speicher fr bettigte Taster

    uint8_t Karten_ID;
    uint8_t Wartezeit_X_10ms;
    uint8_t Karten_ID_Speicherplatz_Zaehler;

    Anzeigedauer=0;
    Wartezeit_X_10ms=0;
    Karten_ID_Speicherplatz_Zaehler=0;
    meinSpezialByte = (0b00000000) | (1<<RFID_MODUL_EIN) | (1<<RFID_ID_NR_EIN) | (1<<RFID_AUS_HEX_EIN);


    #ifdef SERIELL_TRACKING
        Serielle_Anzeige(eeText_Leer_Zeile);
        Serielle_Anzeige(eeText_RFID_Reader);
        Serielle_Anzeige(eeText_Leer_Zeile);
        Serielle_Anzeige(eeText_U2207B_Chip);
        Serielle_Anzeige(eeText_Leer_Zeile);
    #endif


    Display_Anzeige(1, eeText_RFID_Reader);
    Display_Anzeige(2, eeText_U2207B_Chip);


    while(1)
    {

        // ====== Taster 1 ======
        if(!(PIND & (1<<TASTER1)) && (PIND & (1<<TASTER3)))
        {
            if(~meinSpezialByte & (1<<TASTER1))                         // Beim ersten Tastendruck wahr
            {    meinSpezialByte |= (1<<TASTER1);                       // setzt Bit fr Taster1

                Anzeigedauer=0;

                ClearDisplay();

                Display_Anzeige(1, eeText_RFID_Reader);

                if(meinSpezialByte & (1<<RFID_MODUL_EIN))
                {
                    meinSpezialByte &= ~(1<<RFID_MODUL_EIN);
                    PORT_RFID_CFE  &= ~(1<<BIT_RFID_CFE);               // setze Port-Pins auf Low-Pegel

                    Display_Anzeige(2, eeText_Aus);
                }
                else
                {
                    meinSpezialByte |= (1<<RFID_MODUL_EIN);
                    PORT_RFID_CFE |= (1<<BIT_RFID_CFE);                 // setze Port-Pins auf High-Pegel

                    Display_Anzeige(2, eeText_Ein);
                }
            }
        }
        else
            meinSpezialByte &= ~(1<<TASTER1);                           // lscht Bit fr Taster1



        // ====== Taster 2 ======
        if(!(PIND & (1<<TASTER2)))
        {
            if(~meinSpezialByte & (1<<TASTER2))                         // Beim ersten Tastendruck wahr
            {   meinSpezialByte |= (1<<TASTER2);                        // setzt Bit fr Taster2

                Anzeigedauer=0;

                ClearDisplay();

                Display_Anzeige(1, eeText_ID_Nr);

                if(meinSpezialByte & (1<<RFID_ID_NR_EIN))
                {
                    meinSpezialByte &= ~(1<<RFID_ID_NR_EIN);
                    Display_Anzeige(2, eeText_Aus);
                }
                else
                {
                    meinSpezialByte |= (1<<RFID_ID_NR_EIN);
                    Display_Anzeige(2, eeText_Ein);
                }
            }
        }
        else
            meinSpezialByte &= ~(1<<TASTER2);                           // lscht Bit fr Taster2



        // ====== Taster 3 ======
        if(!(PIND & (1<<TASTER3)) && (PIND & (1<<TASTER1)))
        {
            if(~meinSpezialByte & (1<<TASTER3))                         // Beim ersten Tastendruck wahr
            {   meinSpezialByte |= (1<<TASTER3);                        // setzt Bit fr Taster3

                Anzeigedauer=0;

                ClearDisplay();

                Display_Anzeige(1, eeText_Ausgabe);

                if(meinSpezialByte & (1<<RFID_AUS_HEX_EIN))
                {
                    meinSpezialByte &= ~(1<<RFID_AUS_HEX_EIN);
                    Display_Anzeige(2, eeText_Dez);
                }
                else
                {
                    meinSpezialByte |= (1<<RFID_AUS_HEX_EIN);
                    Display_Anzeige(2, eeText_Hex);
                }
            }
        }
        else
            meinSpezialByte &= ~(1<<TASTER3);                           // lscht Bit fr Taster3



        // ====== Taster 1 und 3 ======
        if(!(PIND & (1<<TASTER1)) && !(PIND & (1<<TASTER3)))
        {
            if(~meinSpezialByte & (1<<TASTER1) && ~meinSpezialByte & (1<<TASTER3))        // Beim ersten Tastendruck wahr
            {   meinSpezialByte |= (1<<TASTER1) & (1<<TASTER3);                           // setzt Bit fr Taster2 und Taster3

                Anzeigedauer=0;

                ClearDisplay();

                Display_Anzeige(1, eeText_TAG_Speichern);

                SetDisplayRow(2);
                WriteDisplayString(String);


                OCR1AL = TONHOEHE_GESPEICHERT;                          // Festlegung der Tonhhe im OCR1AL-Register
                TCCR1A = 0b01000000;                                    // Ton an, da COM1A1=0 und COM1A0=1 die Pin-Funktion des Port-Ausgangs fr OC1A Toggle aktivieren


                gesTagKarten.ID = Karten_ID_Speicherplatz_Zaehler;

                eeprom_write_block(&gesTagKarten, &eegesTagKarten[Karten_ID_Speicherplatz_Zaehler], sizeof(gesTagKarten_Struct));


                if(Karten_ID_Speicherplatz_Zaehler<ANZAHL_KARTEN_IDs_IM_EEPROM-1)
                    Karten_ID_Speicherplatz_Zaehler++;
                else
                    Karten_ID_Speicherplatz_Zaehler=0;


                _delay_ms(500);                                         // --- Warte 50 ms ---

                meinSpezialByte |= (1<<RFID_MODUL_EIN);
                PORT_RFID_CFE |= (1<<BIT_RFID_CFE);                     // setze Port-Pins auf High-Pegel

                TCCR1A = 0b00000000;                                    // Ton aus, da COM1A1=0 und COM1A0=0 die Pin-Funktion fr den CTC-Modus deaktivieren
                OCR1AL = TONHOEHE_GELESEN;                              // Festlegung der Tonhhe im OCR1A-Register
            }
        }




        // ====== Relais ausschalten ======
        if((meinSpezialByte & (1<<LED_RELAIS_EIN)) && (Wartezeit_X_10ms==RELAIS_ANZUGSZEIT))
        {
            PORTB &= ~(1<<LED_RELAIS);                                  // LED_RELAIS Low
            meinSpezialByte &= ~(1<<LED_RELAIS_EIN);
        }



        // === Wartezeit die zum Entprellen ===
        // ===      der Tasten dient        ===
        _delay_ms(10);                                                  // --- Warte 10 ms ---

        Wartezeit_X_10ms++;



        if(Gueltige_Daten_empfangen==1)
        {
            Anzeigedauer=0;

            TCCR1A = 0b01000000;                                        // Ton an, da COM1A1=0 und COM1A0=1 die Pin-Funktion des Port-Ausgangs fr OC1A Toggle aktivieren


            // ====== Relais einschalten ======
            for(Karten_ID=0; Karten_ID<ANZAHL_KARTEN_IDs_IM_EEPROM; Karten_ID++)
            {
                eeprom_read_block(&gesTagKarten, &eegesTagKarten[Karten_ID], sizeof(gesTagKarten_Struct));

                for(i=0; i<5; i++)
                    if(RFID_Daten_Byte[i] == gesTagKarten.TagByte[i])
                        meinSpezialByte |= (1<<EEPROM_Tag_Gueltig);
                    else
                    {
                        meinSpezialByte &= ~(1<<EEPROM_Tag_Gueltig);
                        break;
                    }

                if(meinSpezialByte & (1<<EEPROM_Tag_Gueltig))
                {
                    Wartezeit_X_10ms=0;
                    PORTB |= (1<<LED_RELAIS);                           // LED_RELAIS High
                    meinSpezialByte |= (1<<LED_RELAIS_EIN);
                    break;
                }
            }

            Display_Anzeige(1, eeText_Sende);


            if(~meinSpezialByte & (1<<RFID_AUS_HEX_EIN))
                Base=BASE_DEZ;                                          // --- Dez-Darstellung ---
            else
                Base=BASE_HEX;                                          // --- Hex-Darstellung ---

            if(meinSpezialByte & (1<<RFID_ID_NR_EIN))
                AnfangsTagByteNr=0;
            else
                AnfangsTagByteNr=1;


            // Karten-Speicherung frs EEPROM vorbereiten 
            memcpy(gesTagKarten.TagByte, RFID_Daten_Byte, sizeof(RFID_Daten_Byte));


            // Duch das nachfolgende kaskadierte Horner-Schema wird die eingelesene binren
            // Transponderkarten-Nummer in das gewnschte Zahlensystem umgewandelt
            Laenge=10;
            String[Laenge] = '\0';
            Ergebnis = 0b00000000;


            while(Laenge>0) 
            {
                Ergebnis=0;
                TagByteNr=AnfangsTagByteNr;

                while(TagByteNr<5)
                { 
                    i=8;                                                // BinZahlenLaenge = 8

                    while(i>0)
                    { 
                        i--;
                        Ergebnis=(Ergebnis*BASE_BIN);                   // Ergebnis=(Ergebnis*2)

                        if(RFID_Daten_Byte[TagByteNr] & (1<<i))
                            Ergebnis=Ergebnis+1;

                        if(Ergebnis>=Base)                              // Ergebnis>=10
                        {
                            Ergebnis=Ergebnis-Base;                     // Ergebnis=Ergebnis-10
                            RFID_Daten_Byte[TagByteNr] |= (1<<i);
                        }
                        else 
                            RFID_Daten_Byte[TagByteNr] &= ~(1<<i);
                    }

                    TagByteNr++; 
                }

                Laenge--;
                if(Ergebnis<10)
                    String[Laenge] = '0' + Ergebnis;
                else
                    String[Laenge] = 'A' + Ergebnis - 10;
            } 


            SetDisplayRow(2);
            WriteDisplayString(String);

            #ifdef SERIELL_TRACKING
                PutStringInQueueTX(String);
                Serielle_Anzeige(eeText_Leer_Zeile);
            #endif

            _delay_ms(10);                                              // --- Warte 10 ms ---

            TCCR1A = 0b00000000;                                        // Ton aus, da COM1A1=0 und COM1A0=0 die Pin-Funktion fr den CTC-Modus deaktivieren


            Gueltige_Daten_empfangen=0;

            GIMSK   |= (1<<INT1);                                       // Externen Interrupt-INT1-Funktion generell aktivieren
        }
        else
        {
            
            if(meinSpezialByte & (1<<RFID_MODUL_EIN))
            {
                Anzeigedauer++;

                if(Anzeigedauer==255)
                {
                    Anzeigedauer=0;

                    ClearDisplay();

                    Display_Anzeige(1, eeText_Bereit);
                }

            }
        }

    }

}
