/*
/ Projektname : RFID-Reader-Modul
/ Dateiname   : RFID-U2270B.c
/ Version     : 1.0.0 - ATmega8
/ Datum       : 28.12.2013 
/ Compiler    : WinAVR 20100110 libc 1_7_1
/ Autor       : Johannes Kretzmann
/ Copyright   : johannes.kretzmann@projektsammlung.de
/               www.projektsammlung.de
*/





#include <avr/io.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include "RFID-U2270B.h"


/* -------------------------------------------------------------------------------------
   InitRFID-Funktion:
   Parameterbergabe: -
--------------------------------------------------------------------------------------*/
void InitRFID(void)
{

    // ----- Interrupt fr RX_RFID-Schnittstelle definieren ------
    // MCUCR  |= (1<<ISC11) | (1<<ISC10);                           // ISC11=1, ISC10=1, steigende Flanke an INT2 erzeugt einen Interrupt
    // MCUCR  &= ~(1<<ISC10); MCUCR  |= (1<<ISC11);                 // ISC11=1, ISC10=0, fallende Flanke an INT1 erzeugt einen Interrupt
       MCUCR  &= ~(1<<ISC11); MCUCR  |= (1<<ISC10);                 // ISC11=0, ISC10=1, jede Vernderung an INT1 erzeugt einen Interrupt
       GIMSK   |= (1<<INT1);                                        // Externen Interrupt-INT1-Funktion generell aktivieren


    // --- Timer 0 konfigurieren mit CPU-Taktfreqzenz aktivieren ---
    // ------------ Timer0 ist ein 8-Bit Timer (ATMega8) ------------
    // non-PWM mode; Normal port operation; CPU-Clock/64 prescaling
    // TCCR0: -/-/-/-/-/CS02/CS01/CS00
    // 
    // CS02=0,CS01=1,CS00=1                                         // Vorteiler 64
       TCCR0 = 0b00000011;                                          // Timertakt durch Systemtakt und 64-Prescaler erzeugen CS02=0, CS01=1 und CS00=1




    // --- Definition fr den RFID-CFE-Port ---
    PORT_RFID_CFE |= (1<<BIT_RFID_CFE);                             // setze Port-Pins auf High-Pegel
    DDR_RFID_CFE  |= (1<<BIT_RFID_CFE);                             // setzt Port-Pins als Ausgnge fest


    RFID_Synchron_Bit_Zaehler=0;
    RFID_Synchrontakte_aktiv=0;

    Gueltige_Daten_empfangen=0;

    return;
}






/* -------------------------------------------------------------------------------------
   INT1-Interrupt-Funktion: - Externer-Interrupt jede Flankennderung am RFID_RX (INT1/Port-Pin PD3)
                              ausgelst.
                              Ein langer Impuls und ein vorangegangener Low-Pegel  (0) ergeben eine logische 0
                              Ein langer Impuls und ein vorangegangener High-Pegel (1) ergeben eine logische 1
                              Zwei kurze Impuls hintereinander und ein vorangegangener Low-Pegel (0) ergeben eine logische 0
                              Zwei kurze Impuls hintereinander und ein vorangegangener Low-Pegel (1) ergeben eine logische 1

   Parameterbergabe: Parameter werden durch globale Variablen bergeben.

   In dieser Routine wird die Zeit der Intervalle gemessen und in Kurze und Lange Intervalle eingeteilt.
   Zustzlich wird auch noch festgehalten, welchen Pegel der zuletzt detektierte Intervall gehabt hat.

   Parity-Byte: -/-/-/SPALTE1/SPALTE2/SPALTE3/SPLATE4/ZEILENPARITY
--------------------------------------------------------------------------------------*/
ISR(INT1_vect)
{

    uint8_t data;
    uint8_t Pegel;
    uint8_t Parity;
    uint8_t Tag_bits;
    uint8_t Daten_bits;
    uint8_t Synchron_bits;
    uint8_t temp_sreg_INT1;
    uint8_t RFID_Wert_Aktueller_Takt;

    temp_sreg_INT1 = SREG;                                          // Status-Register sichern

    RFID_Wert_Aktueller_Takt  = (uint8_t) TCNT0;
    TCNT0  = 0;                                                     // Timer0 Anfangswert setzen


    // --- Speicherung des Pegel des vorangegangenen Intervalles ---
    if (PIN_RFID_RX & (1<<BIT_RFID_RX))
        Pegel = 1;
    else
        Pegel = 0;


        if(RFID_Wert_Aktueller_Takt>MIN_KURZER_IMPULS && RFID_Wert_Aktueller_Takt<MAX_LANGER_IMPULS)
        {

            Tag_bits   = RFID_Tag_Bit_Zaehler;
            Daten_bits = RFID_Daten_Bit_Zaehler;
            Synchron_bits=RFID_Synchron_Bit_Zaehler;

            Parity = Parity_Byte;

            data = RFID_Tag_Empfangs_Frame;


            if(RFID_Wert_Aktueller_Takt<MIN_LANGER_IMPULS)
            {

                if(RFID_Synchrontakte_aktiv==0)
                {

                    if (Synchron_bits==15)
                    {
                        data=0b00000000;
                        Parity=0b00000000;

                        Tag_bits=0;
                        Daten_bits=0;

                        Vorgaenge_Taktlaenge='S';
                        RFID_Synchrontakte_aktiv=1;
                    }

                    Synchron_bits++;
                }
                else
                {
                    // Zwei kurze Impuls hintereinander ergeben je nach Pegel eine logische 0 oder 1
                    if(Vorgaenge_Taktlaenge=='K')
                        Vorgaenge_Taktlaenge='L';
                    else
                        Vorgaenge_Taktlaenge='K';
                }
            }
            else
            {
                // Ein langer Impuls ergibt je nach Pegel eine logische 0 oder 1
                Vorgaenge_Taktlaenge='L';
                Synchron_bits=0;
            }


            if(RFID_Synchrontakte_aktiv==1 && Vorgaenge_Taktlaenge=='L')
            {

                Tag_bits++;

                // if((Tag_bits % 5)==0 && Tag_bits<=50) 
                if(Tag_bits==5 || Tag_bits==10 ||  Tag_bits==15  || Tag_bits==20 || Tag_bits==25 || Tag_bits==30 || Tag_bits==35 || Tag_bits==40 || Tag_bits==45 || Tag_bits==50)
                {

                    // berprfung des Zeilen-Parity-Bits
                    if(Pegel != (Parity & (1<<ZEILENPARITY)))
                    {
                        Synchron_bits=0;
                        RFID_Synchrontakte_aktiv=0;
                    }

                }
                else
                {

                    data <<= 1;

                    if(Pegel==1)
                    {
                        data |= 0b00000001;

                        // --- Spalten-Parity ---
                        if(Tag_bits<=50)
                            Parity ^= (1<<(4-(Daten_bits % 4)));

                    }
                    else
                        data &= 0b11111110;

                    Daten_bits++;

                    if ((Daten_bits % 8)==0)
                    {
                        RFID_Daten_Byte[(Daten_bits/8)-1] = data;
                        data=0b00000000;
                    }

                }

                // Toggeln des Zeilen-Parity-Bits das das  0-Bit des Parity-Bytes ist
                Parity ^= (0b00000000 | Pegel);                     // Zeilen-Parity
            }



            if (Tag_bits==55)
            {
                Tag_bits=0;
                Synchron_bits=0;
                RFID_Synchrontakte_aktiv=0;

                // --- Das letzte Daten-Byte und das Parity-Byte sind jewels ein Bit nach links verschoben um die ------
                // --- richtigen Werte zu erhalten mssen die Bytes jeweils um ein Bit nach rechts verschoben werden ---
                // --- und die Abfrage auf if(data == (0b00001111 & Parity)) abgendert werden
                // data>>=1;    // Notwendig, da das letzte empfangene RFID-TAG-Byte (5) nur aus 5 Bits besteht
                // Parity>>=1;  // Notwendig, da Datren_bits mit 1 beginnt und somit die Paritybits Bit 1;2;3;4; aber nicht Bit 0 benutzt werden


                // berprfung der Spalten-Parity-Bits
                if(data == (0b00011110 & Parity))
                {
                    Gueltige_Daten_empfangen=1;
                    GIMSK   &= ~(1<<INT1);                          // Externen Interrupt-INT1-Funktion generell deaktivieren
                }

            }


            RFID_Tag_Bit_Zaehler   = Tag_bits;
            RFID_Daten_Bit_Zaehler = Daten_bits;

            RFID_Tag_Empfangs_Frame=data;

            Parity_Byte=Parity;

            RFID_Synchron_Bit_Zaehler = Synchron_bits;

        }
        else
        {
            RFID_Tag_Bit_Zaehler=0;
            RFID_Synchron_Bit_Zaehler=0;

            RFID_Synchrontakte_aktiv=0;
        }


    SREG = temp_sreg_INT1;                                          // Status-Register zurckschreiben
    return;
}



